/*

  stacker.ino - software for an Arduino-based focus stacking and camera controller
  Written by Rylee Isitt
  Rev. 13 Dec 2012
  
  This softare is freeware, although I will not turn away donations ;)
  
*/

#include "nokiaLCD.h"

// ----------------------------------------------------------------------------------------------------
// The values of the variables below can be modified to configure the software for different hardware
// ----------------------------------------------------------------------------------------------------

// pinout configuration
// recommended pinout for the Arduino UNO
#define PIN_IR 2
#define PIN_DATA 3
#define PIN_LATCH 4
#define PIN_CLOCK 5
#define PIN_LCD_SCE   8
#define PIN_LCD_RESET 9
#define PIN_LCD_DC    10
#define PIN_LCD_SDIN  11
#define PIN_LCD_SCLK  12
#define PIN_LCD_LED 13

// configuration for the stepper
const byte stepperMinPeriod = 5; // time (in ms) between steps - fastest speed
const byte stepperMaxPeriod = 20; // time (in ms) between steps - slowest speed
const int micrometersPerHundredHalfSteps = 100; // to figure out this number, advance the platform through a thousand or so half-steps, measure the distance moved with calipers or a micrometer, figure out um/half-step, then multiply by 100 and round to the nearest whole number
boolean halfStepMode = false; // you can toggle this value between true and false to enable or disable half-stepping
boolean reverseMotor = true; // if your stepper is moving in the wrong direction (the Up button move the platform backwards), change this boolean

// configuration for remote
// This project expects the Kenwood RC-P400 remote, available cheaply on eBay from multiple sellers
// If using a different remote, you will have to decipher its protocol and reprogram this sketch
// Not all remotes are equal!
// When choosing a remote, it is best if it is NOT programmable. You want buttons to always produce the same, single code using a single protocol.
// Also make sure any remote you choose has enough buttons. I recommend remotes for CD players, or other remotes designed to control a single device.
// Idea for your own ponderings: adapt this with a radio controller instead
#define REMOTE_INTERRUPT_ON FALLING // can be rising or falling
#define REMOTE_BITS 32 // the number of bits in the IR remote button signal
#define REMOTE_HIGHMAX 2500UL // elapsed time (in microseconds) since last down-sloping IR signal edge above which the code reading will reset
#define REMOTE_HIGHMIN 1500UL // elapsed time (in microseconds) since last down-sloping IR signal edge that acts as a theshhold between a 0 and 1 bit
#define REMOTE_RELEASEDELAY 100000L // elapsed time (in microseconds) after which if no bits have been received we assume that no button is being pressed
// IR codes for buttons
#define REMOTE_BT_NONE 0 // not a real button - a placeholder code for when no buttons are being pressed
#define REMOTE_BT_1 1838317695 // 1
#define REMOTE_BT_2 1838301375 // 2
#define REMOTE_BT_3 1838334015 // 3
#define REMOTE_BT_4 1838293215 // 4
#define REMOTE_BT_5 1838325855 // 5
#define REMOTE_BT_6 1838309535 // 6
#define REMOTE_BT_7 1838342175 // 7
#define REMOTE_BT_8 1838289135 // 8
#define REMOTE_BT_9 1838321775 // 9
#define REMOTE_BT_0 1838285055 // 0
#define REMOTE_BT_ENTER 1838305455 // +10 on Kenwood remote
#define REMOTE_BT_LEFT 1838311575 // Rewind on Kenwood remote
#define REMOTE_BT_RIGHT 1838344215 // Fast Forward on Kenwood remote
#define REMOTE_BT_UP 1838322285 // Volume Up on Kenwood remote
#define REMOTE_BT_DOWN 1838289645 // Volume Down on Kenwood remote
#define REMOTE_BT_STOP 1838319735 // Stop on Kenwood remote
#define REMOTE_BT_PLAYPAUSE 1838336055 // Play/pause on Kenwood remote
#define REMOTE_BT_MODE 1838348295 // P. Mode on Kenwood remote
#define REMOTE_BT_HALFPRESS 1838285820 // AI AUTO on Kenwood remote
#define REMOTE_BT_FULLPRESS 1838302140 // AI TIMER on Kenwood remote
#define REMOTE_BT_BACKLIGHT 1838318460 // PRESENCE on Kenwood remote

// shift register bitmasks
// If you change the pinout for the shift register, just modify the constants below
const byte stepperEnableA = B00000001; // shift register output pin QA/Q0
const byte stepperEnableB = B00000010; // shift register output pin QB/Q1
const byte stepperA1 = B00000100; // shift register output pin QC/Q2
const byte stepperA2 = B00001000; // shift register output pin QD/Q3
const byte stepperB1 = B00010000; // shift register output pin QE/Q4
const byte stepperB2 = B00100000; // shift register output pin QF/Q5
const byte shutterHalf = B01000000; // shift register output pin QG/Q6
const byte shutterFull = B10000000; // shift register output pin QH/Q7

// ----------------------------------------------------------------------------------------------------
// Stepper motor phase states - no need to modify these unless you want to use different excitation modes 
// ----------------------------------------------------------------------------------------------------

// two-phase on, full step coil states
//   A1 B1 A2 B2
// 1 1  1  0  0
// 2 1  1  0  0
// 3 0  1  1  0
// 4 0  1  1  0
// 5 0  0  1  1
// 6 0  0  1  1
// 7 1  0  0  1
// 8 1  0  0  1
// at peak current all of the time, more heat, more torque
// makes the most out of a weak motor
const byte coilA1StateFull = B11000011;
const byte coilA2StateFull = B00111100;
const byte coilB1StateFull = B11110000;
const byte coilB2StateFull = B00001111;

// half-step coil states
//   A1 B1 A2 B2
// 1 1  0  0  0
// 2 1  1  0  0
// 3 0  1  0  0
// 4 0  1  1  0
// 5 0  0  1  0
// 6 0  0  1  1
// 7 0  0  0  1
// 8 1  0  0  1
// at peak current 1/2 of the time, less heat, less torque
// good setting for more powerful motors
const byte coilA1StateHalf = B11000001;
const byte coilA2StateHalf = B00011100;
const byte coilB1StateHalf = B01110000;
const byte coilB2StateHalf = B00000111;

// ----------------------------------------------------------------------------------------------------
// Tracking variables - no need to modify these unless you are changing the software code
// ----------------------------------------------------------------------------------------------------

// setup LCD
nokiaLCD lcd;

// these variables set/track the state of the shift register
byte registerState = B00000000;
boolean updateRegister = true;

// variables for getting IR codes
// if using a remote that uses 16 bits or less for its code
// change the variable type of code and lastCode to unsigned int
volatile unsigned long code = REMOTE_BT_NONE;
volatile unsigned long lastCode = REMOTE_BT_NONE;
volatile byte bitCount = 0;
volatile unsigned long lastBitTimestamp = 0;
volatile unsigned long curTime = 0;
volatile unsigned long elapsed = 0;

// defines for tracking button states
// this is a bit messy, but saves memory as opposed to using booleans (which are actually bytes instead of bits!)
#define BT_NONE 0
#define BT_1_DN 1
#define BT_1_UP 2
#define BT_2_DN 3
#define BT_2_UP 4
#define BT_3_DN 5
#define BT_3_UP 6
#define BT_4_DN 7
#define BT_4_UP 8
#define BT_5_DN 9
#define BT_5_UP 10
#define BT_6_DN 11
#define BT_6_UP 12
#define BT_7_DN 13
#define BT_7_UP 14
#define BT_8_DN 15
#define BT_8_UP 16
#define BT_9_DN 17
#define BT_9_UP 18
#define BT_0_DN 19
#define BT_0_UP 20
#define BT_ENTER_DN 21
#define BT_ENTER_UP 22
#define BT_LEFT_DN 23
#define BT_LEFT_UP 24
#define BT_RIGHT_DN 25
#define BT_RIGHT_UP 26
#define BT_UP_DN 27
#define BT_UP_UP 28
#define BT_DOWN_DN 29
#define BT_DOWN_UP 30
#define BT_STOP_DN 31
#define BT_STOP_UP 32
#define BT_PLAYPAUSE_DN 33
#define BT_PLAYPAUSE_UP 34
#define BT_MODE_DN 35
#define BT_MODE_UP 36
#define BT_HALFPRESS_DN 37
#define BT_HALFPRESS_UP 38
#define BT_FULLPRESS_DN 39
#define BT_FULLPRESS_UP 40
#define BT_BACKLIGHT_DN 41
#define BT_BACKLIGHT_UP 42

byte buttonState = BT_NONE;

// variables for stepping
long stepperPos = 0; // measured in half steps, even if using full step mode
long startPos = 0; // measured in half steps, even if using full step mode
long endPos = 0; // measured in half steps, even if using full step mode

// ----------------------------------------------------------------------------------------------------
// This function is called by an interrupt when the IR receiver breakout board signal falls
// It measures the intervals between falling edges and uses these intervals to decode the IR signal
// ----------------------------------------------------------------------------------------------------

void irFire() {
  /*
    Deciphering codes from the remote works by watching the elapsed time between down-sloping edges (as a high signal transitions to a low signal)
    When this happens, irFire() is triggered by an interrupt
    ```|___|````|____
       a        b
    In the waveform above, this elapsed time is the period between "a" and "b"
    If the elapsed time is between REMOTE_HIGHMIN and REMOTE_HIGHMAX, a 1 is shifted left into the code variable
    If the elapsed time is below REMOTE_HIGHMIN, a 0 is shifted left into the code variable
    If the elapsed time is above REMOTE_HIGHMAX, this resets code reading (sets code to 0 and bitCount to 0)
    Once REMOTE_BITS bits have been shifted in, the code is completely received and stored in the lastCode variable
  */

  curTime = micros();
  elapsed = curTime - lastBitTimestamp;
  lastBitTimestamp = curTime;

  // watch the bits coming from the receiver
  bitCount++;
  if (elapsed < REMOTE_HIGHMIN) { // zero
    code <<= 1;
  } else if (elapsed < REMOTE_HIGHMAX) { // one
    code <<= 1;
    code++;
  } else { // begin a new code
    code = 0;
    bitCount = 0;
  }

  if (bitCount == REMOTE_BITS) { // a complete code has been received received
    lastCode = code;
    bitCount = 0;
    code = 0;
  }

}

// ----------------------------------------------------------------------------------------------------
// This function sends the current register state to the shift register
// ----------------------------------------------------------------------------------------------------

void updateShiftRegister() {
  // update the shift register
  if (updateRegister) {
    digitalWrite(PIN_LATCH, LOW);
    shiftOut(PIN_DATA, PIN_CLOCK, MSBFIRST, registerState);
    digitalWrite(PIN_LATCH, HIGH);
    updateRegister = false;
  }
}

// ----------------------------------------------------------------------------------------------------
// This function sets the stepperEnableA bit in the shift register state
// ----------------------------------------------------------------------------------------------------

void setStepperEnableA (boolean setting) {
  if (setting && ((registerState & stepperEnableA) == 0)) {
    registerState |= stepperEnableA;
    updateRegister = true;
  } else if (!setting && ((registerState & stepperEnableA) > 0)) {
    registerState &= ~stepperEnableA;
    updateRegister = true;
  }
}

// ----------------------------------------------------------------------------------------------------
// This function sets the stepperEnableB bit in the shift register state
// ----------------------------------------------------------------------------------------------------

void setStepperEnableB (boolean setting) {
  if (setting && ((registerState & stepperEnableB) == 0)) {
    registerState |= stepperEnableB;
    updateRegister = true;
  } else if (!setting && ((registerState & stepperEnableB) > 0)) {
    registerState &= ~stepperEnableB;
    updateRegister = true;
  }
}

// ----------------------------------------------------------------------------------------------------
// This function sets the stepperA1 bit in the shift register state
// ----------------------------------------------------------------------------------------------------

void setStepperA1 (boolean setting) {
  if (setting && ((registerState & stepperA1) == 0)) {
    registerState |= stepperA1;
    updateRegister = true;
  } else if (!setting && ((registerState & stepperA1) > 0)) {
    registerState &= ~stepperA1;
    updateRegister = true;
  }
}

// ----------------------------------------------------------------------------------------------------
// This function sets the stepperA2 bit in the shift register state
// ----------------------------------------------------------------------------------------------------

void setStepperA2 (boolean setting) {
  if (setting && ((registerState & stepperA2) == 0)) {
    registerState |= stepperA2;
    updateRegister = true;
  } else if (!setting && ((registerState & stepperA2) > 0)) {
    registerState &= ~stepperA2;
    updateRegister = true;
  }
}

// ----------------------------------------------------------------------------------------------------
// This function sets the stepperB1 bit in the shift register state
// ----------------------------------------------------------------------------------------------------

void setStepperB1 (boolean setting) {
  if (setting && ((registerState & stepperB1) == 0)) {
    registerState |= stepperB1;
    updateRegister = true;
  } else if (!setting && ((registerState & stepperB1) > 0)) {
    registerState &= ~stepperB1;
    updateRegister = true;
  }
}

// ----------------------------------------------------------------------------------------------------
// This function sets the stepperB2 bit in the shift register state
// ----------------------------------------------------------------------------------------------------

void setStepperB2 (boolean setting) {
  if (setting && ((registerState & stepperB2) == 0)) {
    registerState |= stepperB2;
    updateRegister = true;
  } else if (!setting && ((registerState & stepperB2) > 0)) {
    registerState &= ~stepperB2;
    updateRegister = true;
  }
}

// ----------------------------------------------------------------------------------------------------
// This function sets the shutterHalf bit in the shift register state
// ----------------------------------------------------------------------------------------------------

void setShutterHalf (boolean setting) {
  if (setting && ((registerState & shutterHalf) == 0)) {
    registerState |= shutterHalf;
    updateRegister = true;
  } else if (!setting && ((registerState & shutterHalf) > 0)) {
    registerState &= ~shutterHalf;
    updateRegister = true;
  }
}

// ----------------------------------------------------------------------------------------------------
// This function sets the shutterFull bit in the shift register state
// ----------------------------------------------------------------------------------------------------

void setShutterFull (boolean setting) {
  if (setting && ((registerState & shutterFull) == 0)) {
    registerState |= shutterFull;
    updateRegister = true;
  } else if (!setting && ((registerState & shutterFull) > 0)) {
    registerState &= ~shutterFull;
    updateRegister = true;
  }
}

// ----------------------------------------------------------------------------------------------------
// This function moves the stepper by one step (or one half-step) in either direction
// ----------------------------------------------------------------------------------------------------

void singleStep(boolean forward, boolean halfMode) {

  static byte stepperStageBitMask = B10000000;

  // move stepperStageBitMask forwards or backwards
  if ((forward && !reverseMotor) || (!forward && reverseMotor)) {
    stepperStageBitMask = stepperStageBitMask >> 1;
    if (stepperStageBitMask == 0) stepperStageBitMask = B10000000;
    stepperPos++;
  } else {
    stepperStageBitMask = stepperStageBitMask << 1;
    if (stepperStageBitMask == 0) stepperStageBitMask = B00000001;
    stepperPos--;
  }
  
  if (halfMode) {
    setStepperEnableA(((stepperStageBitMask & coilA1StateHalf) > 0) || ((stepperStageBitMask & coilA2StateHalf) > 0));
    setStepperEnableB(((stepperStageBitMask & coilB1StateHalf) > 0) || ((stepperStageBitMask & coilB2StateHalf) > 0));
    setStepperA1((stepperStageBitMask & coilA1StateHalf) > 0);
    setStepperA2((stepperStageBitMask & coilA2StateHalf) > 0);
    setStepperB1((stepperStageBitMask & coilB1StateHalf) > 0);
    setStepperB2((stepperStageBitMask & coilB2StateHalf) > 0);
  } else {
    setStepperEnableA(((stepperStageBitMask & coilA1StateFull) > 0) || ((stepperStageBitMask & coilA2StateFull) > 0));
    setStepperEnableB(((stepperStageBitMask & coilB1StateFull) > 0) || ((stepperStageBitMask & coilB2StateFull) > 0));
    setStepperA1((stepperStageBitMask & coilA1StateFull) > 0);
    setStepperA2((stepperStageBitMask & coilA2StateFull) > 0);
    setStepperB1((stepperStageBitMask & coilB1StateFull) > 0);
    setStepperB2((stepperStageBitMask & coilB2StateFull) > 0);
  }

}

// ----------------------------------------------------------------------------------------------------
// When continuously stepping, this function checks to see if it's time to step again, and if so, makes
// a single step (or half-step)
// ----------------------------------------------------------------------------------------------------

void multiStep(boolean forward, boolean halfMode) {

  static byte stepPeriod = stepperMaxPeriod; // ms per step (current speed)
  static unsigned long nextStepTime = 0;

  unsigned long curMillis = millis();
  if (curMillis >= nextStepTime) {
    if (stepPeriod > stepperMinPeriod) stepPeriod--;
    if ((curMillis - nextStepTime) > 100) stepPeriod = stepperMaxPeriod;
    singleStep(forward, halfMode);
    nextStepTime = curMillis + stepPeriod;
  }

}

// ----------------------------------------------------------------------------------------------------
// This function rotates the stepper to a destination position in either full-step or half-step mode
// ----------------------------------------------------------------------------------------------------

void gotoStep(int goalPos, boolean halfMode) {

  // go to goalPos
  while (stepperPos != goalPos) {
    if (((stepperPos < goalPos) && !reverseMotor) || ((stepperPos > goalPos) && reverseMotor)) {
      multiStep(true,halfMode);
      updateShiftRegister();
    } else {
      multiStep(false,halfMode);
      updateShiftRegister();
    }
  }

}

// ----------------------------------------------------------------------------------------------------
// This function sets the button state based on the received IR codes and the timing of the last bit
// ----------------------------------------------------------------------------------------------------

void handleButtons(void) {

  switch (lastCode) {
    case REMOTE_BT_1:
      buttonState = BT_1_DN;
      break;
    case REMOTE_BT_2:
      buttonState = BT_2_DN;
      break;
    case REMOTE_BT_3:
      buttonState = BT_3_DN;
      break;
    case REMOTE_BT_4:
      buttonState = BT_4_DN;
      break;
    case REMOTE_BT_5:
      buttonState = BT_5_DN;
      break;
    case REMOTE_BT_6:
      buttonState = BT_6_DN;
      break;
    case REMOTE_BT_7:
      buttonState = BT_7_DN;
      break;
    case REMOTE_BT_8:
      buttonState = BT_8_DN;
      break;
    case REMOTE_BT_9:
      buttonState = BT_9_DN;
      break;
    case REMOTE_BT_0:
      buttonState = BT_0_DN;
      break;
    case REMOTE_BT_ENTER:
      buttonState = BT_ENTER_DN;
      break;
    case REMOTE_BT_LEFT:
      buttonState = BT_LEFT_DN;
      break;
    case REMOTE_BT_RIGHT:
      buttonState = BT_RIGHT_DN;
      break;
    case REMOTE_BT_UP:
      buttonState = BT_UP_DN;
      break;
    case REMOTE_BT_DOWN:
      buttonState = BT_DOWN_DN;
      break;
    case REMOTE_BT_STOP:
      buttonState = BT_STOP_DN;
      break;
    case REMOTE_BT_PLAYPAUSE:
      buttonState = BT_PLAYPAUSE_DN;
      break;
    case REMOTE_BT_MODE:
      buttonState = BT_MODE_DN;
      break;
    case REMOTE_BT_HALFPRESS:
      buttonState = BT_HALFPRESS_DN;
      break;
    case REMOTE_BT_FULLPRESS:
      buttonState = BT_FULLPRESS_DN;
      break;
    case REMOTE_BT_BACKLIGHT:
      buttonState = BT_BACKLIGHT_DN;
      break;
    case REMOTE_BT_NONE:
      if (buttonState == BT_1_DN) {
        buttonState = BT_1_UP;
      } else if (buttonState == BT_2_DN) {
        buttonState = BT_2_UP;
      } else if (buttonState == BT_3_DN) {
        buttonState = BT_3_UP;
      } else if (buttonState == BT_4_DN) {
        buttonState = BT_4_UP;
      } else if (buttonState == BT_5_DN) {
        buttonState = BT_5_UP;
      } else if (buttonState == BT_6_DN) {
        buttonState = BT_6_UP;
      } else if (buttonState == BT_7_DN) {
        buttonState = BT_7_UP;
      } else if (buttonState == BT_8_DN) {
        buttonState = BT_8_UP;
      } else if (buttonState == BT_9_DN) {
        buttonState = BT_9_UP;
      } else if (buttonState == BT_0_DN) {
        buttonState = BT_0_UP;
      } else if (buttonState == BT_ENTER_DN) {
        buttonState = BT_ENTER_UP;
      } else if (buttonState == BT_LEFT_DN) {
        buttonState = BT_LEFT_UP;
      } else if (buttonState == BT_RIGHT_DN) {
        buttonState = BT_RIGHT_UP;
      } else if (buttonState == BT_UP_DN) {
        buttonState = BT_UP_UP;
      } else if (buttonState == BT_DOWN_DN) {
        buttonState = BT_DOWN_UP;
      } else if (buttonState == BT_STOP_DN) {
        buttonState = BT_STOP_UP;
      } else if (buttonState == BT_PLAYPAUSE_DN) {
        buttonState = BT_PLAYPAUSE_UP;
      } else if (buttonState == BT_MODE_DN) {
        buttonState = BT_MODE_UP;
      } else if (buttonState == BT_HALFPRESS_DN) {
        buttonState = BT_HALFPRESS_UP;
      } else if (buttonState == BT_FULLPRESS_DN) {
        buttonState = BT_FULLPRESS_UP;
      } else if (buttonState == BT_BACKLIGHT_DN) {
        buttonState = BT_BACKLIGHT_UP;
      } else {
        buttonState = BT_NONE;
      }
      break;
  }

  /*
    it takes time for IR codes to be completely received
    and if no button is being pressed, the interrupt irFire won't trigger
    so the timing for deciding if the remote is inactive must be done outside of the interrupt
    we can't turn interrupts off without interfering with getting IR codes and the interrupts can interfere with timing
    we must assume that the interrupt will fire WHILE the following code is executing, leading to situations where curTime - lastBitTimestamp may be negative
    so we use a signed long to calculate this difference
    note that this bit of code also enables tracking of the button down state, even though the remote only outputs a button's code when it is initially pressed
    this works because the remote continues to output "keep alive" bits while the button is held
  */
  if (lastCode != REMOTE_BT_NONE) {
    long bitDelayElapsed = micros() - lastBitTimestamp;
    if (bitDelayElapsed > REMOTE_RELEASEDELAY) {
      lastCode = REMOTE_BT_NONE;
    }
  }

}

// ----------------------------------------------------------------------------------------------------
// This function adds digits to various configuration parameters in the UI
// ----------------------------------------------------------------------------------------------------

unsigned int addDigit(unsigned int value, byte digit) {
  const unsigned int maxValue = 65535;

  if (value <= (maxValue / 10)) {
    value *= 10;
    if (digit > 0) {
      if (value <= (maxValue - digit)) {
        value += digit;
      } else {
        value = maxValue;
      }
    }
  } else {
    value = maxValue;
  }

  return value;
}

// ----------------------------------------------------------------------------------------------------
// This function handles the UI, stepping, and camera control in the stacking mode
// ----------------------------------------------------------------------------------------------------

void stackMode(void) {

  // paramVals[0]: step count
  // paramVals[1]: wait time (milliseconds between slices - the default of 15000 ms is to prevent flash overheating)
  // paramVals[2]: bracket count (number of shots to take per slice)
  // paramVals[3]: time (in milliseconds) that shutter release is held down
  #define STACKPARAMS_SIZE 4
  static unsigned int paramVals[STACKPARAMS_SIZE] = {80, 15000, 1, 100};
  const unsigned int paramMin[STACKPARAMS_SIZE] = {1, 0, 1, 20};
  const char paramNames[STACKPARAMS_SIZE][4] = {"Slc", "Dly", "Bkt", "Exp"};
  const char paramSuffix[STACKPARAMS_SIZE][4] = {"", "ms", "", "ms"};
  static byte curParam = 0;
  static boolean editField = false;

  if (buttonState == BT_UP_UP) { // cycle selected field up
    if (curParam > 0) {
      curParam--;
    } else {
      curParam = STACKPARAMS_SIZE - 1;
    }
    editField = false;
  } else if (buttonState == BT_DOWN_UP) { // cycle selected field down
    curParam++;
    if (curParam >= STACKPARAMS_SIZE) curParam = 0;
    editField = false;
  } else if (buttonState == BT_ENTER_UP) { // toggle field editing
    editField = !editField;
  } else if (buttonState == BT_LEFT_UP) { // remove the last digit from the value of the selected field
    paramVals[curParam] /= 10;
    editField = true;
  } else if (buttonState == BT_0_UP) { // add a 0 to the end of the value of the selected field
    paramVals[curParam] = addDigit(paramVals[curParam], 0);
    editField = true;
  } else if (buttonState == BT_1_UP) { // add a 1 to the end of the value of the selected field
    paramVals[curParam] = addDigit(paramVals[curParam], 1);
    editField = true;
  } else if (buttonState == BT_2_UP) { // add a 2 to the end of the value of the selected field
    paramVals[curParam] = addDigit(paramVals[curParam], 2);
    editField = true;
  } else if (buttonState == BT_3_UP) { // add a 3 to the end of the value of the selected field
    paramVals[curParam] = addDigit(paramVals[curParam], 3);
    editField = true;
  } else if (buttonState == BT_4_UP) { // add a 4 to the end of the value of the selected field
    paramVals[curParam] = addDigit(paramVals[curParam], 4);
    editField = true;
  } else if (buttonState == BT_5_UP) { // add a 5 to the end of the value of the selected field
    paramVals[curParam] = addDigit(paramVals[curParam], 5);
    editField = true;
  } else if (buttonState == BT_6_UP) { // add a 6 to the end of the value of the selected field
    paramVals[curParam] = addDigit(paramVals[curParam], 6);
    editField = true;
  } else if (buttonState == BT_7_UP) { // add a 7 to the end of the value of the selected field
    paramVals[curParam] = addDigit(paramVals[curParam], 7);
    editField = true;
  } else if (buttonState == BT_8_UP) { // add a 8 to the end of the value of the selected field
    paramVals[curParam] = addDigit(paramVals[curParam], 8);
    editField = true;
  } else if (buttonState == BT_9_UP) { // add a 9 to the end of the value of the selected field
    paramVals[curParam] = addDigit(paramVals[curParam], 9);
    editField = true;
  } else if (buttonState == BT_PLAYPAUSE_UP) { // do the stack!
      // make sure values are not below minimums
      for (byte i = 0; (i < STACKPARAMS_SIZE); i++) {
         if (paramVals[i] < paramMin[i]) paramVals[i] = paramMin[i];
      }

      // update display
      lcd.clearText();
      lcd.putText("--Stacking--",0,0,true);
      lcd.putText("Press stop",1,0,false);
      lcd.putText("To cancel.",2,0,false);
      lcd.putText("!\"\"\"\"\"\"\"\"\"\"#",4,0,false,true);
      lcd.updateText();
      
      // stack
      for (int i=0; i<paramVals[0]; i++) {
        unsigned int waitHalfTime = (paramVals[1] + paramVals[3]) / 2;
        
        // calculate target position
        int gotoPos = startPos + ((long(i)*(long(endPos-startPos))) / (int(paramVals[0]) - 1));
        
        // update progress bar
        byte progress = ((i*12) / (paramVals[0]-1));
        for (byte j=0; j<progress; j++) {
          if (j == 0) {
            lcd.putText("$",4,j,false,true);
          } else if (j == 11) {
            lcd.putText("&",4,j,false,true);
          } else {
            lcd.putText("%",4,j,false,true);
          }
        }
        lcd.updateText();
        
        // go to target position
        gotoStep(gotoPos, halfStepMode);
        
        // take picture
        for (byte n=0; n<paramVals[2]; n++) {
        
          // enable half-press mode
          setShutterHalf(true);
          updateShiftRegister();
          
          delay(waitHalfTime); // wait (to reduce camera vibrations)

          // trigger camera shutter
          setShutterFull(true);
          updateShiftRegister();
          delay(paramVals[3]);
          
          // release camera shutter
          setShutterHalf(false);
          setShutterFull(false);
          updateShiftRegister();

          delay(waitHalfTime); // wait (to reduce camera vibrations)
          
        }
       
        // if stop button pressed, cancel
        handleButtons();
        if ((buttonState == BT_STOP_UP) || (buttonState == BT_STOP_DN)) i = paramVals[0];
        
      }
      
  } else {

    // make sure values are not below minimums
    if (!editField) {
      for (byte i = 0; (i < STACKPARAMS_SIZE); i++) {
         if (paramVals[i] < paramMin[i]) paramVals[i] = paramMin[i];
      }
    }

    // update display
    lcd.clearText();
    lcd.putText("Stack Setup-",0,0,true);
    for (byte i = 0; i < STACKPARAMS_SIZE; i++) {
      if (i == curParam) {
        if (editField) {
          lcd.putText(String(paramNames[i])+">"+String(paramVals[i])+"_",i+1,0);
        } else {
          lcd.putText(String(paramNames[i])+">"+String(paramVals[i])+String(paramSuffix[i]),i+1,0);
        }
      } else {
        lcd.putText(String(paramNames[i])+":"+String(paramVals[i])+String(paramSuffix[i]),i+1,0);
      }
    }
    lcd.updateText();

  }

}

// ----------------------------------------------------------------------------------------------------
// This function handles the UI and stepping in the position mode
// ----------------------------------------------------------------------------------------------------

void positionMode(void) {

  // mode 1: positioning - allows manual movement of rack and setting of extents
  //    up: moves rack forward
  //    down: moves rack backward
  //    left: goes to start position
  //    right: goes to end position
  //    play/pause: sets start position
  //    stop: sets end position
  //    repeat: toggles continuous vs incremental movement

  // deal with setting of start and end points
  if (buttonState == BT_PLAYPAUSE_UP) {
    startPos = stepperPos;
  } else if (buttonState == BT_STOP_UP) {
    endPos = stepperPos;
  } else if (buttonState == BT_UP_DN) {
    multiStep(true, halfStepMode);
  } else if (buttonState == BT_DOWN_DN) {
    multiStep(false, halfStepMode);
  } else if (buttonState == BT_RIGHT_UP) {
    gotoStep(endPos, halfStepMode);
  } else if (buttonState == BT_LEFT_UP) {
    gotoStep(startPos, halfStepMode);
  } else { // update display
  
    // calculate physical distancce between start and end points
    unsigned int stepDiff = abs(startPos - endPos);
    unsigned long distance = ((long)stepDiff * (long)micrometersPerHundredHalfSteps) / 100L;
    
    lcd.clearText();
    lcd.putText("--Position--",0,0,true);
    lcd.putText("Pos:"+String(stepperPos),1,0);
    lcd.putText("Stt:"+String(startPos),2,0);
    lcd.putText("End:"+String(endPos),3,0);
    lcd.putText("Dist:"+String(distance)+"um",4,0);
    lcd.updateText();
  }

}

// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------

void setup(void) {

  Serial.begin(9600);
  lcd.begin(PIN_LCD_SCE, PIN_LCD_RESET, PIN_LCD_DC, PIN_LCD_SDIN, PIN_LCD_SCLK, PIN_LCD_LED);

  // setup pin modes for shift register
  pinMode(PIN_DATA, OUTPUT);
  pinMode(PIN_CLOCK, OUTPUT);
  pinMode(PIN_LATCH, OUTPUT);
  updateShiftRegister();

  // setup pin for IR in
  pinMode(PIN_IR, INPUT);
  digitalWrite(PIN_IR, HIGH); // pulls high
  attachInterrupt(0, irFire, REMOTE_INTERRUPT_ON);
  interrupts();

}

// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------

void loop(void) {

  // handle interpretation of codes received from remote
  handleButtons();

  /*
  mode button cycles through the modes below:
  mode 1: positioning - allows manual movement of rack and setting of extents
     up: moves rack forward, while button is held if enableButtonHold is true, otherwise until start/stop is pressed
     down: moves rack backward, while button is held if enableButtonHold is true, otherwise until start/stop is pressed
     start/stop: stops rack motion
     left: goes to start position
     right: goes to end position
     enter: sets current position as an extent
         followed by left: sets start extent
         followed by right: sets end extent
  mode 2: stacking - automatically increments between start and stop positions while triggering camera shutter
     up: cycles up through parameters
     down: cycles down through parameters
     enter: toggles editing of parameters
         up: increments numeric parameter by 1
         down: decrements numeric parameter by 1
         left: divides numeric parameter by 2
         right: multiplies numeric parameter by 2
     start/stop: starts or stops stacking
  mode 3: interval - for timelapse, with support for a motorized mount driven by a stepper (WORK IN PROGRESS)
  mode 4: equatorial - for driving a barn door mount using a stepper and doing long exposures while mount is running (WORK IN PROGRESS)
  */

  // mode toggle
  static byte mode = 1;
  if (buttonState == BT_MODE_UP) {
    mode++;
    if (mode > 2) mode = 1;
  }

  // enable/disable backlight
  if (buttonState == BT_BACKLIGHT_UP) lcd.toggleBacklight();

  // shutter
  static boolean halfPressState = false;
  if (buttonState == BT_HALFPRESS_UP) {
    halfPressState = !halfPressState;
    setShutterHalf(halfPressState);
  }
  if (buttonState == BT_FULLPRESS_DN) {
    setShutterFull(true);
  } else {
    setShutterFull(false);
  }

  // mode state
  if (mode == 1) {
    positionMode();
  } else if (mode == 2) {
    stackMode();
  }

  // update the register
  updateShiftRegister();

}
